#include "troxnet.h"
#define DEBUG(x) 

#define TROXPORT 12777 // the portnumber Trox uses


typedef struct {
    char * name;
    struct sockaddr_in id;
    int sock;
} Host;

typedef struct {
    int spieler;
    int a;
    Vector n;
    bool collision;
} Message;

typedef struct {
    int spieler;
    int a;
} Shortmessage;

int hosts;
Host host[4];
int localsock;
int localplayers;
int allplayers;


void closenet(void)
{
    int i;
    DEBUG(printf("troxnet: Shutting down...\n");)
    close(localsock);
    for (i=0; i<hosts; i++) {
	close(host[i].sock);
    }
}

void initport(void)
{
    struct sockaddr_in id;
    localsock = socket(AF_INET, SOCK_DGRAM,0);
    if (localsock<0) {
	perror("troxnet: error opening datagram socket!\n");
	exit(1);
    }
    id.sin_family = AF_INET;
    id.sin_addr.s_addr = INADDR_ANY;
    id.sin_port = htons(TROXPORT);
    if (bind(localsock,(struct sockaddr *) &id,sizeof(id))) {
	perror("troxnet: error binding socket!\n");
	exit(1);
    }
}


void initsockets(void)
{
    int i;
    struct hostent * hp;
    DEBUG(printf("troxnet: Initializing send-sockets...\n");)
    for (i=0; i<hosts;i++) {
	DEBUG(printf("troxnet: Socket to %s.\n",host[i].name);)
	host[i].sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (host[i].sock <0) {perror("troxnet: error opening send-socket");exit(1);}
	if (fcntl(host[i].sock,F_SETFL,O_NONBLOCK)<0) {
	    perror("troxnet: error setting nonblocking mode"); 
	    exit(1);
	}
	hp=gethostbyname(host[i].name);
	if (hp==0) {printf("troxnet: %s: unknown host!\n",host[i].name);exit(1);}
	bcopy(hp->h_addr, &host[i].id.sin_addr, hp->h_length);
	host[i].id.sin_family = AF_INET;
	host[i].id.sin_port = htons(TROXPORT);
    }
}


void sendmessage(const int & spieler, const signed int & a, const Vector &n, const bool &collision)
{
    int i;
    Message message;
    message.spieler=spieler;
    message.a=a;
    message.n=n;
    message.collision=collision;
    for (i=0; i<hosts; i++) {
	if (sendto(host[i].sock,(void *)&message,sizeof(message),0,
		   (sockaddr*)&host[i].id,sizeof(host[i].id))<0) {
	    perror("Error sending message");
	    exit(1);
	}
    }
}


bool sendshortmessage(const int & spieler, const int & a)
{
    int i;
    Shortmessage message;
    message.spieler=spieler;
    message.a=a;
    for (i=0; i<hosts; i++) {
	if (sendto(host[i].sock,(void *)&message,sizeof(message),
		   0,(sockaddr*)&host[i].id,sizeof(host[i].id))<0) {
	    return false;
	}
    }
    return true;
}


void getmessage(int &spieler, int &a, Vector & n, bool &collision, bool lock)
{
    Message message;
    if (lock) {fcntl(localsock,F_SETFL,0);}
    else {fcntl(localsock,F_SETFL,O_NONBLOCK);}
    while (read(localsock,&message,sizeof(message))==-1);
    spieler=message.spieler;
    a=message.a;
    n=message.n;
    collision=message.collision;
}


bool getshortmessage(int &spieler, int &a, bool lock)
{
    Shortmessage message;
    if (lock) {fcntl(localsock,F_SETFL,0);}
    else {fcntl(localsock,F_SETFL,O_NONBLOCK);}
    if (read(localsock,&message,sizeof(message))==-1) {return false;}
    spieler=message.spieler;
    a=message.a;
    return true;
}


void connectplayers(Player * p)
{
    int readycount=0;
    int i,a;
    printf("Connecting (press ESC-RETURN to abort) ...\n");
    do {
	while (getshortmessage(i,a,false)) {
	    if (!p[i].alive) {printf("Player %d ready!\n",i+1); p[i].alive=true;readycount++;}
	}
	sleep(1);
	for (i=0; i<allplayers; i++) {
	    if (vga_getkey()==27) {printf("Program aborted by user!\n");exit(0);}
	    if (p[i].local) {sendshortmessage(i,0);}
	}
    }
    while (readycount<allplayers);
}


int initnet(int argc,char * argv[], int &ap, Player * p)
{
    int i,k;
    int found;
    DEBUG(printf("troxnet: Initialising connections...\n");)
    if (argc<3) {
	printf("troxnet: Less than 2 players specified!\nTry: strox localhost localhost\n"); 
	exit(1);
    }
    if (argc>5) {
	printf("troxnet: More than 4 players specified!\n"); 
	exit(1);
    }
    localplayers=0;
    for (i=1; i<argc; i++) {
	if (strcmp(argv[i],"localhost")==0) {localplayers++;}
    }
    if (localplayers<1) {printf("troxnet: No player at localhost!\n"); exit(1);}
    if (localplayers>2) {printf("troxnet: More than 2 players at localhost!\n"); exit(1);}
    allplayers=argc-1;
    DEBUG(printf("troxnet: Got %d players\n",allplayers);)
    DEBUG(printf("troxnet: Got %d player(s) on localhost.\n",localplayers);)
    hosts=0; 
    for (i=1;i<argc;i++) {
	found=0;
	for (k=0;k<i-1;k++) {
	    if (strcmp(host[k].name,argv[i])==0) {found=1;}
	}
	if (!found) {
	    if ((host[hosts].name=(char*)malloc(strlen(argv[i])*sizeof(char)))==NULL) {
		printf("troxnet: could not get enough memory!\n"); 
		exit(1);
	    }
	    strcpy(host[hosts].name,argv[i]);
	    hosts++;
	}
    }
    DEBUG(printf("troxnet: Got %d different host(s).\n",hosts);)
    DEBUG(printf("troxnet: Initializing TROX-port...\n");)
    initport();
    k=0;
    for (i=0; i<allplayers; i++) {
	p[i].local=(strcmp(argv[i+1],"localhost")==0);
	p[i].score=0;
	p[i].alive=false;
	if (p[i].local) {p[i].side=k; k++;}
	printf("Player %d (",i+1);
	if (i==0) {printf("red");}
	if (i==1) {printf("green");}
	if (i==2) {printf("yellow");}
	if (i==3) {printf("orange");}
	printf(") on %s ",argv[i+1]);
	if (p[i].local) {
	    if (p[i].side==0) {
		printf("/ left side ");
	    } 
	    else {
		printf("/ right side ");
	    }
	}
	printf("\n");
    }
    initsockets();
    atexit(closenet);
    connectplayers(p);
    ap=allplayers;
    return argc-1;
}
